#*****************************************************************************
#
#    Sage extension class "MTX", based on the Aachen Meataxe
#
#    Copyright (C) 2009 Simon A. King <simon.king@uni-jena.de>
#
#  Distributed under the terms of the GNU General Public License (GPL),
#  version 2 or later (at your choice)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************
"""
C-MeatAxe Matrices

This module is a wrapper for a modified version of the Aachen
`C-MeatAxe <http://www.math.rwth-aachen.de/homes/MTX/download.html>`_ 
version  2.2.4 and provides matrices over the finite field `\\mathbb F_q`, 
where `q\\le 255`.

C-MeatAxe matrices can be read from a file named ``f`` via ``M=MTX(f)``. 
If ``f==''``, an empty matrix is created.
If ``q`` is a prime power, ``M=MTX(q,[[a11,a12,...,],...,[...,amn]])`` 
creates a matrix over `\\mathbb F_q` with the given marks. 

AUTHORS:

- Simon King <simon.king@uni-jena.de>

"""

####################
# 
# import sage types
#
####################

import sage
import sage.all
from warnings import showwarning
import sys
from sage.all import Integer
from sage.all import cputime
from sage.all import GF
from sage.all import is_prime_power
from sage.all import factor
from sage.all import save
from sage.all import cached_method, cached_function
from sage.all import SAGE_ROOT

## The tables for MeatAxe computations will be found in the package data folder
import os
os.environ['MTXLIB'] = SAGE_ROOT + '/local/pGroupCohomology/mtx_tables/'


#########################################################################
#########################################################################
# 
# Definition of the MTX extension type (declarations in mtx.pxd)
# The Data-attribute is already cdefined in mtx.pxd
# 
#########################################################################
#########################################################################

class MTX_unpickle_class:
    """
    Unpickling :class:`~pGroupCohomology.mtx.MTX` matrices

    EXAMPLES::

        sage: from pGroupCohomology.mtx import MTX
        sage: M=MTX(5,[[1,2,3,4]])
        sage: M == loads(dumps(M))   # indirect doctest
        True

    """
    def __init__(self):
        """
        Unpickling :class:`~pGroupCohomology.mtx.MTX` matrices

        TESTS::

            sage: from pGroupCohomology.mtx import MTX_unpickle_class, MTX
            sage: unpickle = MTX_unpickle_class()   # indirect doctest
            sage: M=MTX(5,[[1,2,3,4]])
            sage: L=M.__reduce__()[1]
            sage: N=unpickle(*L)
            sage: print N
            [1 2 3 4]

        """
        self.__safe_for_unpickling__=True

    def __call__(self,f,nr,nc,object Data,m):
        """
        Unpickling :class:`~pGroupCohomology.mtx.MTX` matrices

        TESTS::

            sage: from pGroupCohomology.mtx import MTX_unpickle_class, MTX
            sage: unpickle = MTX_unpickle_class()
            sage: M=MTX(5,[[1,2,3,4]])
            sage: L=M.__reduce__()[1]
            sage: N=unpickle(*L)   # indirect doctest
            sage: print N
            [1 2 3 4]

        """
        cdef MTX OUT
        cdef char* x
        #cdef list L
        cdef int i
        cdef int NR
        cdef PTR p
        if isinstance(Data,list) and Data:
            Lnext = Data.__iter__().next
            NR = nr
            OUT = MTX(f,nr,nc,mutable=m)
            p = OUT.Data.d
            for i from 0 <= i < NR:
                x=PyString_AsString(Lnext())
                zstr2row(p,x)
                zsteprow(&(p))
            return OUT
        if Data:
            x=PyString_AsString(Data)
            OUT = MTX(f,nr,nc,mutable=m)
            str2mat(OUT.Data,len(Data),x)
            return OUT
        else:
            return MTX('',mutable=m)

mtx_unpickle = MTX_unpickle_class()


cdef class MTX:
    """
MTX-Matrix extension class

INPUT:

MTX matrices can be created by providing one of the following input data,
where ``q<255`` is a prime power, ``m,n`` are non-negative integers, and
``a11,...,amn`` are integers.

- '', the empty string => empty matrix over an unspecified field
- a string ``f`` ==> load matrix from the file named ``f``
- ``q, [[a11,a12,...a1n],...,[am1,...]]``
  ==> `m \\times n` matrix over ``GF(q)``, filled with ``a11,a12,...``
- ``q, m, n`` ==> zero `m \\times n` matrix over ``GF(q)``
- ``q, m, "id"`` ==> unit `m \\times m` matrix over ``GF(q)``


If the optional parameter ``mutable`` is ``False`` (by default, it is ``True``), the resulting
matrix can not be changed, and it can be used as key in a Python dictionary.

NOTE:

If ``q`` is not a prime power, a ``MemoryError`` will be raised. 

EXAMPLES::

    sage: from pGroupCohomology.mtx import MTX

1. Creating an empty matrix::

    sage: M = MTX('')
    sage: M
    Empty MTX matrix
    sage: print M
    []

2. Creating a zero (3x2)-matrix::

    sage: M = MTX(4,3,2)
    sage: M
    (3 x 2) MTX matrix over GF(4)
    sage: print M
    [0 0]
    [0 0]
    [0 0]

3. Creating a matrix from a list of lists::

    sage: M = MTX(5,[[1,2,3],[4,5,6]])
    sage: M
    (2 x 3) MTX matrix over GF(5)
    sage: print M
    [1 2 3]
    [4 0 1]

4. Creating a unit matrix::

    sage: M = MTX(3,2,'id')
    sage: M
    (2 x 2) MTX matrix over GF(3)
    sage: print M
    [1 0]
    [0 1]

5. Creating a matrix from a file.

   First, we have to create that file; we use a temporary file, that will be removed
   when leaving Sage. Note that the method :meth:`msave` must be used, which
   does not use Python pickling but relies on the intrinsic C--MeatAxe way of
   saving.
   ::

    sage: f = tmp_filename()
    sage: M.msave(f)
    sage: N = MTX(f)
    sage: print N
    [1 0]
    [0 1]
    """

## See mtx.pxd:
##    cdef matrix_t *Data

###################
## Init, Dealloc, Copy
    def __init__(self, f,Dt=[],Dx=1,mutable=True):
        """
        MTX-Matrix extension class

        INPUT:

        MTX matrices can be created by providing one of the following input data,
        where ``q<255`` is a prime power, ``m,n`` are non-negative integers, and
        ``a11,...,amn`` are integers.

        - '', the empty string => empty matrix over an unspecified field
        - a string ``f`` ==> load matrix from the file named ``f``
        - ``q, [[a11,a12,...a1n],...,[am1,...]]``
          ==> `m \\times n` matrix over ``GF(q)``, filled with ``a11,a12,...``
        - ``q, m, n`` ==> zero `m \\times n` matrix over ``GF(q)``
        - ``q, m, "id"`` ==> unit `m \\times m` matrix over ``GF(q)``


        If the optional parameter ``mutable`` is ``False`` (by default, it is ``True``), the resulting
        matrix can not be changed, and it can be used as key in a Python dictionary.

        NOTE:

        If ``q`` is not a prime power, a ``MemoryError`` will be raised. 

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX

        1. Creating an empty matrix::

            sage: M = MTX('')
            sage: M
            Empty MTX matrix
            sage: print M
            []

        2. Creating a zero (3x2)-matrix::

            sage: M = MTX(4,3,2)
            sage: M
            (3 x 2) MTX matrix over GF(4)
            sage: print M
            [0 0]
            [0 0]
            [0 0]

        3. Creating a matrix from a list of lists::

            sage: M = MTX(5,[[1,2,3],[4,5,6]])  # indirect doctest
            sage: M
            (2 x 3) MTX matrix over GF(5)
            sage: print M
            [1 2 3]
            [4 0 1]

        4. Creating a unit matrix::

            sage: M = MTX(3,2,'id')
            sage: M
            (2 x 2) MTX matrix over GF(3)
            sage: print M
            [1 0]
            [0 1]

        5. Creating a matrix from a file.

           First, we have to create that file; we use a temporary file, that will be removed
           when leaving Sage. Note that the method :meth:`msave` must be used, which
           does not use Python pickling but relies on the intrinsic C--MeatAxe way of
           saving.
           ::

            sage: f = tmp_filename()
            sage: M.msave(f)
            sage: N = MTX(f)
            sage: print N
            [1 0]
            [0 1]
            
        """
        cdef int i,j
        cdef long ff 
        self._mutable=mutable
        if self.Data:
            matfree(self.Data)
        if isinstance(f,basestring):
            if (f==''):
                return
            else:
                f = os.path.realpath(f)
                try:
                    fsock = open(f,"rb",0)
                    fsock.close()
                except (OSError,IOError):
                    return
                self.Data = matload(f)
                if zsetfield(self.Data.fl):
                    raise ValueError, "Invalid saved date"
            return
        if (isinstance(f,Integer) or isinstance(f,int) or isinstance(f,long)):
            ff = f
            if (isinstance(Dt,list) or isinstance(Dt,tuple)) and \
                   (len(Dt)>=0) and (len(Dt[0])>=0):
                #zsetfield(ff)
                #zsetlen(len(Dt[0]))
                self.Data = matalloc(f, len(Dt), len(Dt[0]))
            elif ((isinstance(Dt,int)) or isinstance(Dt,Integer)) and \
                 ((isinstance(Dx,int)) or isinstance(Dx,Integer)) and \
                 ((Dx>=0) and (Dt>=0)):
                #zsetfield(ff)
                #zsetlen(Dx)
                self.Data = matalloc(ff, Dt,Dx)
                if not self.Data:
                    raise MemoryError, "Error allocating MTX matrix"
                return
            elif ((isinstance(Dt,int)) or isinstance(Dt,Integer) or isinstance(Dt,long)) and \
                 (Dx=="id"):
                if zsetfield(ff):
                    raise ValueError, "Field order must be a prime power < 256"
                self.Data = MTXmatid(ff, <long>Dt)
                if not self.Data:
                    raise MemoryError, "Error allocating MTX matrix"
                return
            else:
                raise TypeError, "Invalid definition of matrix elements"
        cdef PTR x,y
        if not self.Data:
            raise MemoryError, "Error allocating MTX matrix"
        x = self.Data.d
        cdef int nr = self.Data.nor
        cdef int nc = self.Data.noc
        cdef int idx
        cdef list dt = Dt
        for i from 0 <= i < nr:
            y = x
            idx = 0
            dtnext = dt[i].__iter__().next
            for j from 0 <= j < nc:
                # if Dt[i][j]>=0:
                zinsert_step(&(y), &(idx), zitof(dtnext()))#dt[i][j]))
                # else:
                #     zinsert(x, j+1, taddinv[abs(Dt[i][j])%f])
            zsteprow(&(x))
                            
    def __dealloc__(self):
        """
        Deallocate MTX matrix

        TESTS::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(5,[[1,2,3,4],[4,3,2,1]])
            sage: del M    # indirect doctest
            sage: M
            Traceback (most recent call last):
            ...
            NameError: name 'M' is not defined
        """
        if self.Data:
            matfree(self.Data)
        cdef matrix_t *p
        self.Data = p

    def __copy__(self):
        """
        Return a copy of an MTX matrix

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M = MTX(2,[20*[0],20*[0],[1]+19*[0]])
            sage: N = copy(M)    # indirect doctest
            sage: print N
            [0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0]
            [0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0]
            [1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0]
            sage: N==M
            True
            sage: N is M
            False
            sage: M=MTX('')
            sage: N=copy(M)
            sage: N
            Empty MTX matrix
            sage: N==M
            True
            sage: N is M
            False
        """        
        cdef MTX retval
        retval = MTX('',mutable=self._mutable)
        if self.Data:
            retval.Data = matdup(self.Data)
            if not retval.Data:
                raise MemoryError, "Error copying a MTX matrix"
        return(retval)
    
    def __reduce__(self):
        """
        Pickle/save an MTX matrix.

        TESTS::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(5,[[1,2,3,4]])
            sage: M.__reduce__()[0] 
            <pGroupCohomology.mtx.MTX_unpickle_class instance at ...>
            sage: M.__reduce__()[1]
            (5, 1, 4, '&d', True)
            sage: print loads(dumps(M))
            [1 2 3 4]
        """
        cdef char* d
        cdef L = []
        Lnext = L.append
        cdef int i,NR
        cdef PTR p
        if self.Data:
            zsetfield(self.Data.fl)
            zsetlen(self.Data.noc)
            #d = mat2str(self.Data)
            #return mtx_unpickle, (self.Data.fl,self.Data.nor,self.Data.noc,PyString_FromStringAndSize(d,zsize(self.Data.nor)),self._mutable)
            NR = self.Data.nor
            p = self.Data.d
            for i from 0 <= i < NR:
                d = zrow2str(p)
                Lnext(PyString_FromStringAndSize(d,zrowsize_io))
                zsteprow(&(p))
            return mtx_unpickle, (self.Data.fl,self.Data.nor,self.Data.noc,''.join(L),self._mutable) # or ...,L,self._mutable)
        else:
            return mtx_unpickle, (0,0,0,'',self._mutable)

##################
## ==,<,>, dictionary
    def __richcmp__(MTX self,other,x):
        """
        Compare two MTX matrices
        
        Of course, '<' and '>' doesn't make much sense for matrices.
        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(2,[20*[0],20*[0],[1]+19*[0]])
            sage: N=copy(M)
            sage: M==N   # indirect doctest
            True
            sage: M!=N
            False
            sage: print M<N
            None
            sage: N[2,19]=1
            sage: M==N
            False
            sage: M!=N
            True

        TESTS::

            sage: M == int(1)
            False

        """
        cdef MTX N
        # < 0, <= 1, == 2, != 3, > 4, >= 5
        try:
            N = other
        except TypeError:
            if x==3:
                return True
            if x==2:
                return False
            return None
        if N is None:
            if x==3:
                return True
            if x==2:
                return False
            return None
        cdef char* d1
        cdef char* d2
        if x==3:
            if not (self.Data):
                if not (N.Data):
                    return False
                else:
                    return True
            if not (N.Data):
                return True
            #zsetfield(self.Data.fl)
            #zsetlen(self.Data.noc)
            #d1 = mat2str(self.Data)
            #d2 = mat2str(N.Data)
            #return ((self.Data.fl!=N.Data.fl) or (self.Data.noc!=N.Data.noc) or (self.Data.nor!=N.Data.nor) or \
            #    PyString_FromStringAndSize(d1,zsize(self.Data.nor)) != PyString_FromStringAndSize(d2,zsize(N.Data.nor)))
            return (matcmp(self.Data,N.Data)==0)
        if (x==2):
            if not (self.Data):
                if not(N.Data):
                    return True
                else:
                    return False
            if not (N.Data):
                return False
            #zsetfield(self.Data.fl)
            #zsetlen(self.Data.noc)
            #d1 = mat2str(self.Data)
            #d2 = mat2str(N.Data)
            #return (self.Data.fl == N.Data.fl) and (self.Data.noc==N.Data.noc) and (self.Data.nor==N.Data.nor) and \
            #       PyString_FromStringAndSize(d1,zsize(self.Data.nor)) == PyString_FromStringAndSize(d2,zsize(N.Data.nor))
            return (matcmp(self.Data,N.Data)==1)
        if (x==1 or x==5):
            if not (self.Data):
                if not(N.Data):
                    return True
                else:
                    return False
            #zsetfield(self.Data.fl)
            #zsetlen(self.Data.noc)
            #d1 = mat2str(self.Data)
            #d2 = mat2str(N.Data)
            #if (self.Data.fl == N.Data.fl) and (self.Data.noc==N.Data.noc) and (self.Data.nor==N.Data.nor) and \
            #       PyString_FromStringAndSize(d1,zsize(self.Data.nor)) == PyString_FromStringAndSize(d2,zsize(N.Data.nor)):
            #    return True
            if matcmp(self.Data,N.Data):
                return True
        return None

    def is_mutable(self):
        """
        Tell whether self is mutable

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(5,[[1,2,3,4]])
            sage: M.is_mutable()
            True
            sage: N=MTX(5,[[1,2,3,4]],mutable=False)
            sage: N.is_mutable()
            False
        """
        return self._mutable

    def is_immutable(self):
        """
        Tell whether self is immutable

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(5,[[1,2,3,4]])
            sage: M.is_immutable()
            False
            sage: N=MTX(5,[[1,2,3,4]],mutable=False)
            sage: N.is_immutable()
            True
        """
        return not self._mutable

    def set_immutable(self):
        """
        Turn self immutable, so that it can be hashed and used as dictionary key

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(5,[[1,2,3,4]])
            sage: N=MTX(5,[[4,3,2,1]],mutable=False)
            sage: D={N:2}
            sage: D[M]=1
            Traceback (most recent call last):
            ...
            AttributeError: Mutable objects are not hashable
            sage: M.set_immutable()
            sage: D[M]=1
            sage: D[N]
            2
            sage: D[copy(M)]
            1
        """
        self._mutable=False

    def set_mutable(self):
        """
        Turn self mutable, so that it can be changed

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: N=MTX(5,[[4,3,2,1]],mutable=False)
            sage: N[0,1]=0
            Traceback (most recent call last):
            ...
            ValueError: Matrix is immutable
            sage: N.set_mutable()
            sage: N[0,1]=0
            sage: print N
            [4 0 2 1]
        """
        self._mutable=True

    def __hash__(self):
        """
        M.__hash__(): Return a hash value for M, provided M is immutable

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M = MTX(3, [[randint(0,2) for _ in range(5)] for __ in range(5)])
            sage: M.set_immutable()
            sage: hash(M) == hash(M._memory_string())
            True
            sage: hash(M) == hash(copy(M))   # indirect doctest
            True

        """
        cdef char* d
        if self._mutable:
            raise AttributeError, "Mutable objects are not hashable"
        else:
            if not self.Data:
                return 0
            else:
                zsetfield(self.Data.fl)
                zsetlen(self.Data.noc)
                d = mat2str(self.Data)
                return hash(PyString_FromStringAndSize(d,zsize(self.Data.nor)))

    def _memory_string(self):
        """
        Return a Python string corresponding to self's memory block.

        Note:

        This method is mainly for creating a consistency test for the hash.

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M = MTX(3, [[randint(0,2) for _ in range(5)] for __ in range(5)])
            sage: M.set_immutable()
            sage: hash(M) == hash(M._memory_string())
            True
            sage: type(M._memory_string())
            <type 'str'>
 
        """
        cdef char* d
        if not self.Data:
            return ""
        else:
            zsetfield(self.Data.fl)
            zsetlen(self.Data.noc)
            d = mat2str(self.Data)
            return PyString_FromStringAndSize(d,zsize(self.Data.nor))

##########################
## Saving should be done via pickling
## However, we kept a method that relies on MeatAxe matsave:
    def msave(self,f):
        """
        Save self into file, using the original C--MeatAxe format
     
        INPUT:

        ``f`` - a file name (string)

        NOTE: 

        If ``M`` is an MTX matrix than it can be saved in the usual way
        by ``save(M,'filename')`` and reloaded with ``load('filename')``.
        Saving with ``M.msave('filename')`` uses a different format, so that
        the data can be reloaded with ``MTX('filename')``.

        EXAMPLES:

        First, we create a temporary file name, that will be removed
        when leaving Sage. Of course, in an application a permanent
        file would usually be chosen.
        ::
        
            sage: f = tmp_filename(ext='.sobj')

        Next, we create a matrix and use msave::
        
            sage: from pGroupCohomology.mtx import MTX
            sage: M = MTX(5,[[1,2,3,4],[4,3,2,1]])
            sage: M.msave(f)

        It can not be reloaded with 'load'::
        
            sage: N = load(f)
            Traceback (most recent call last):
            ...
            UnpicklingError: ...

        But it can be reloaded with the MTX constructor::
        
            sage: N = MTX(f)
            sage: M==N
            True
        """
        matsave(self.Data,f)

##########################
## Structural parts

    cpdef int nrows(self):
        """
        Number of rows

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(5,[[4,3,2,1],[3,2,1,0]])
            sage: M.nrows()
            2
            sage: M=MTX('') # the empty matrix
            sage: M.nrows()
            0
        """
        if self.Data:
            return(self.Data.nor)
        else:
            return 0
    cpdef int ncols(self):
        """
        Number of columns

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(5,[[4,3,2,1],[3,2,1,0]])
            sage: M.ncols()
            4
            sage: M=MTX('') # the empty matrix
            sage: M.ncols()
            0
        """
        if self.Data:
            return(self.Data.noc)
        else:
            return 0

    cpdef int base(self) except -1:
        """
        Return the order of the field over which the matrix is defined

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(25,[[4,3,2,1]])
            sage: M.base()
            25
        """
        if self.Data:
            return self.Data.fl
        raise ValueError, "%s has no base"%repr(self)
 
    cpdef int characteristic(self) except -1:
        """
        Return the characteristic of the field over which the matrix is defined

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(25,[[4,3,2,1]])
            sage: M.characteristic()
            5
        """
        if self.Data:
            return factor(self.Data.fl)[0][0]
        raise ValueError, "%s has no characteristic"%repr(self)

    def __getitem__(self,key):
        """
        Return element, row, or slice of self.

        INPUT:
            key -- tuple (i,j) with i, j non-negative integers
         or key -- non-negative integer
         or key -- slice object, created via [i:j]

        USAGE:
            A[i, j] -- the i,j of A, or
            A[i]    -- the i-th row of A, or
            A[i:j]  -- the i-th through (j-1)-st rows of A.

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(9,[[1,2,3],[4,5,6],[7,8,9]])
            sage: print M
            [1 2 3]
            [4 5 6]
            [7 8 0]
            sage: M[1,2]    # indirect doctest
            6
            sage: print M[1]
            [4 5 6]
            sage: print M[1:3]
            [4 5 6]
            [7 8 0]
        """
        if not self.Data:
            raise IndexError, "Matrix is empty"
        zsetfield(self.Data.fl)
        cdef matrix_t *mtxrow
        cdef PTR p
        cdef MTX A
        cdef int i,j
        cdef int nr=self.Data.nor
        cdef int nc=self.Data.noc
        zsetlen(nc)
        if isinstance(key,tuple):
            # key is a tuple, so we get i and j efficiently, construct
            # corresponding integer entry.
            if (len(key) != 2):
                raise IndexError, "index must be an integer or a pair of integers"
            i = key[0]
            j = key[1]
            if (i<0) or (i>=nr) or (j<0) or (j>=nc):
                raise IndexError, "index out of range"
            else:
                p = self.Data.d
                zadvance(&(p),i)
                #print i
                #print &(p)==&(self.Data.d)
                return zftoi(zextract(p,j+1))
        elif isinstance(key, slice):
            # Slice interpretation is passed to the sequence
            # constructed by range
            if (key.start>=0) and (key.stop<=nr) and \
               (key.start<=nr) and (key.stop>=0):
                A=MTX('',self._mutable)
                A.Data = _matextract(self.Data, long(key.start+1), long(key.stop-key.start))
                if A.Data:
                    return A
                else:
                    raise IndexError, "index out of range"
            else:
                raise IndexError, "index out of range"
        elif (isinstance(key,sage.rings.integer.Integer) or \
              isinstance(key,int)) and (key>=0) and (key<nr):
            A=MTX('',self._mutable)
            A.Data = _matextract(self.Data, long(key+1), long(1))
            if A.Data:
                return A
            else:
                raise IndexError, "index out of range"
        else:
            raise IndexError, "index out of range"

    def __setitem__(self, key, el):
        """
        x.__setitem__(key, el) <==> x[key]=el

        USAGE:
            i) key= pair of integers, el= integer
           ii) key= integer, el= list/tuple of integers 
               row number <key> is filled with the entries of <el>,
               until the end of the row or the end of <el> is reached
          iii) key= integer, el= MTX with the same number of columns as self
               The rows of self starting with number <key> are filled with the 
               contents of el.

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(9,[[1,2,3],[4,5,6],[7,8,9]])
            sage: print M
            [1 2 3]
            [4 5 6]
            [7 8 0]
            sage: M[0]=[3,2,1]   # indirect doctest
            sage: print M
            [3 2 1]
            [4 5 6]
            [7 8 0]
            sage: M[1,2]=0
            sage: print M
            [3 2 1]
            [4 5 0]
            [7 8 0]
            sage: N=MTX(9,[[1,2,3],[3,4,5]])
            sage: M[1]=N
            sage: print M
            [3 2 1]
            [1 2 3]
            [3 4 5]
        """
        if not self._mutable:
            raise ValueError, "Matrix is immutable"
        cdef PTR x
        cdef int j,mini,idx
        cdef MTX tmp
        #cdef list L
        if self.Data:
            zsetfield(self.Data.fl)
        else:
            raise IndexError, "Matrix is empty"
        if (isinstance(key,tuple) or isinstance(key,list)) and \
               (len(key)==2) and (isinstance(key[0],int) or isinstance(key[0],Integer)) and \
               (isinstance(key[1],int) or isinstance(key[1],Integer)):
            if (key[0]>=0) and (key[0]<self.Data.nor) and (key[1]>=0) and (key[1]<self.Data.noc):
                x = self.Data.d
                zsetlen(self.Data.noc)
                zadvance(&(x),key[0])
                zinsert(x, key[1]+1, zitof(el))
            else:
                raise IndexError, "index out of range"
        elif (isinstance(key,Integer) or isinstance(key,int)): # insert a list or another MTX
            if (key>=0) and (key<self.Data.nor):
                x = self.Data.d
                zsetlen(self.Data.noc)
                zadvance(&(x),key)
                if isinstance(el,MTX):
                    tmp = el
                    if tmp.Data.fl!=self.Data.fl:
                        raise ArithmeticError, "Fields are incompatible"
                    if tmp.Data.nor+key>self.Data.nor:
                        raise IndexError, "MTX to be inserted has too many rows"
                    if tmp.Data.noc!=self.Data.noc:
                        raise IndexError, "MTX to be inserted must have the same number of columns as self"
                    zsetlen(tmp.Data.noc)
                    zcpyrow(x,tmp.Data.d,zsize(tmp.Data.nor))
                    zsetlen(self.Data.noc)
                    return
                L = el.__iter__()
                Lnext = L.next
                mini = min(len(el), self.Data.noc)
                idx = 0
                for j from 0 <= j < mini:
                    zinsert_step(&(x), &(idx), zitof(Lnext())) #[j]))
            else:
                raise IndexError, "index out of range"
        else:
            raise IndexError, "index must be an integer or a pair of integers"

    def _rowlist_(self, i, j=-1):
        """
        M._rowlist_(i): Return row <i> as a list
        M._rowlist_(i,j): Return rows <i> up to and including <j> in a single list
        
        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(9,[[1,2,3],[4,5,6],[7,8,9]])
            sage: M._rowlist_(0)
            [1, 2, 3]
            sage: M._rowlist_(1,2)
            [4, 5, 6, 7, 8, 0]
        """
        if self.Data:
            zsetfield(self.Data.fl)
            zsetlen(self.Data.noc)
        else:
            raise TypeError, "Matrix is empty"
        if not(isinstance(i,int) or isinstance(i,Integer)):
            raise TypeError, "Index must be an integer"
        if (i<0) or (i>=self.Data.nor):
            raise IndexError, "Index out of range"
        cdef PTR p,q
        cdef int k,l,ii,jj, idx
        p = self.Data.d
        zadvance(&(p),i)
        idx = 0
        q = p
        cdef list L = [zftoi(zextract_step(&(q),&(idx))) for k in range(self.Data.noc)]
        if j!=-1:
            if not(isinstance(j,int) or isinstance(j,Integer)):
                raise TypeError, "Second index must be an integer"
            if j >= self.Data.nor:
                raise IndexError, "Index out of range"
            ii=i
            jj=j
            for k from ii < k <= jj:
                zsteprow(&(p))
                idx = 0
                q = p
                for l from 0<=l<self.Data.noc:
                    L.append(zftoi(zextract_step(&(q),&(idx))))
        return L

    cpdef list _matlist_(self):
        """
        M._matlist_(): Return the entries of M as a list of lists

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(9,[[1,2,3],[4,5,6],[7,8,9]])
            sage: M._matlist_()
            [[1, 2, 3], [4, 5, 6], [7, 8, 0]]
        """
        if self.Data:
            zsetfield(self.Data.fl)
            zsetlen(self.Data.noc)
        else:
            raise IndexError, "Matrix is empty"
        cdef PTR p,q
        cdef int i,j
        p = self.Data.d
        cdef int idx
        cdef list l_out=[]
        for i from 1<=i<self.Data.nor:
            q = p
            idx = 0
            l_out.append([zftoi(zextract_step(&(q),&(idx))) for j in range(self.Data.noc)])
            zsteprow(&(p))
        idx = 0
        l_out.append([zftoi(zextract_step(&(p),&(idx))) for j in range(self.Data.noc)])
        return l_out

#########################
## Arithmetics
    def __add__(MTX self, MTX right):
        """
        Add two MTX matrices of equal size

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(5,[[4,3,2,1],[3,2,1,0]])
            sage: N=MTX(5,[[1,1,2,2],[3,3,4,4]])
            sage: print M+N    # indirect doctest
            [0 4 4 3]
            [1 0 0 4]
        """
        if ((self.Data.noc==0) or (right.Data.noc==0)):
            raise NotImplementedError, "The matrices must not be empty"
        cdef MTX left
        left = self.__copy__()
        if not right._mutable:
            left.set_immutable()
        if matadd(left.Data,right.Data):
            return left
        else:
            raise ArithmeticError, "Matrix sizes or fields not compatible"

    def __sub__(MTX self, MTX right):
        """
        Subtract two MTX matrices of equal size

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(5,[[4,3,2,1],[3,2,1,0]])
            sage: N=MTX(5,[[1,1,2,2],[3,3,4,4]])
            sage: print M-N   # indirect doctest
            [3 2 0 4]
            [0 4 2 1]
        """
        if (self.Data.fl==2):
            return self.__add__(right)
        if ((self.Data.noc==0) or (right.Data.noc==0)):
            raise NotImplementedError, "The matrices must not be empty"
        cdef MTX left
        left = self.__copy__()
        if right.is_immutable():
            left.set_immutable()
        if matsub(left.Data,right.Data):
            return left
        else:
            raise ArithmeticError, "Matrix sizes or fields not compatible"

    def __neg__(self):
        """
        Negation of a MTX matrix

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(5,[[4,3,2,1],[3,2,1,0]])
            sage: print -M   # indirect doctest
            [1 2 3 4]
            [2 3 4 0]
        """
        if (self.Data.fl==2):
            return self.__copy__()
        if (self.ncols()==0):
            raise NotImplementedError, "The matrix must not be empty"
        cdef MTX result
        result = self.__copy__()
        if matneg(result.Data):
            return result
        else:
            raise ArithmeticError, "Something went wrong"

    def __mul__(MTX self, right):
        """
        Multiplication of MTX matrices

        NOTE:

        -M is different from M*(-1), if the field is not of prime order,
        since then -1 does not represent the additive inverse of 1.

        NOTE:

        We only implemented the multiplication matrix times integer,
        but not integer times matrix.

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(9,[[1,2,3],[4,5,6],[7,8,9]])
            sage: print -M
            [2 1 6]
            [8 7 3]
            [5 4 0]
            sage: print M*(-1)   # indirect doctest
            [8 4 5]
            [1 6 7]
            [3 2 0]
            sage: print M*6
            [6 3 8]
            [5 2 4]
            [1 7 0]
            sage: N=MTX(9,[[2,3,4],[5,6,7],[8,9,0]])
            sage: print M*N   # indirect doctest
            [2 6 6]
            [2 6 3]
            [2 6 0]
            sage: print N*M
            [3 0 5]
            [3 0 5]
            [8 4 5]

        """
        if isinstance(right,MTX):
            return self._mulMTX_(right)
        elif isinstance(right,int) or isinstance(right,Integer):
            return self._mulInt_(right)

    cpdef MTX _mulMTX_(self, MTX right):
        """
Multiply two MTX matrices of equal size

NOTE:
Usually, this method would only be called by MTX.__mul__

EXAMPLES::

    sage: from pGroupCohomology.mtx import MTX
    sage: M=MTX(9,[[1,2,3],[4,5,6],[7,8,9]])
    sage: N=MTX(9,[[2,3,4],[5,6,7],[8,9,0]])
    sage: print M._mulMTX_(N)
    [2 6 6]
    [2 6 3]
    [2 6 0]
        """
        if ((self.ncols()==0) or (right.ncols()==0)):
            raise ValueError, "The matrices must not be empty"
        if self.Data.fl != right.Data.fl:
            raise ValueError, "Different base fields"
        #cdef MTX left = self.__copy__()
        cdef MTX left = MTX(self.Data.fl, self.Data.nor, right.Data.noc, mutable=(self._mutable or right._mutable))
        
        #if right.is_immutable():
        #    left.set_immutable()
        #if matmul(left.Data,right.Data):
        #    if not self.Data.nor:
        #        print 'self\n', self.Data.nor, self.Data.noc
        #        print 'right\n', right
        #        print 'left\n',left.Data.nor, left.Data.noc
        #    return left
        #else:
        #    raise ArithmeticError, "Matrix sizes or fields not compatible"
        r = matmul_result(self.Data, right.Data, left.Data) 
        if r == 0:
            return left
        else:
            raise ValueError, "Matrix sizes are not compatible"

    cpdef MTX _mulInt_(self, long right):
        """
multiply an MTX matrix with a field element represented by an integer

NOTE:
Normally, this method would only be called by MTX.__mul__

EXAMPLES::

    sage: from pGroupCohomology.mtx import MTX
    sage: M=MTX(9,[[1,2,3],[4,5,6],[7,8,9]])
    sage: print M._mulInt_(6)
    [6 3 8]
    [5 2 4]
    [1 7 0]
    sage: print M._mulInt_(-1)
    [8 4 5]
    [1 6 7]
    [3 2 0]
        """
        cdef MTX left
        if right==0:
            return MTX(self.Data.fl, self.Data.nor, self.Data.noc, mutable=self._mutable) # this is a zero matrix
        elif right==1:
            return self.__copy__()
        if (self.ncols()==0):
            raise NotImplementedError, "The matrix must not be empty"
        left = self.__copy__()
        if matmulF(left.Data,zitof(right)):
            return left
        else:
            raise ArithmeticError, "Matrix sizes or fields not compatible"
            
    def __div__(MTX self, long p):
        """
        Divide an MTX matrix by a field element represented by an integer

        NOTE:
            If the field is not of prime order then the integer -1 does not
            represent the additive inverse of 1.

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(9,[[1,2,3],[4,5,6],[7,8,9]])
            sage: print M/3    # indirect doctest
            [5 7 1]
            [3 8 2]
            [4 6 0]
        """
        if ((p%self.Data.fl)==0):
            raise ZeroDivisionError, "Cannot divide by zero"
        elif p==1:
            return self.__copy__()
        elif not self.Data:
            return self
        return self._mulInt_(tmultinv[zitof(p)])

    def __pow__(MTX self,n,m):
        """
        M.__pow__(n): return M^n

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(9,[[2,3,4],[5,6,7],[8,9,0]])
            sage: print M^4   # indirect doctest
            [6 0 1]
            [2 2 7]
            [4 4 1]
        """
        if (self.ncols()==0):
            raise NotImplementedError, "The matrix must not be empty"
        cdef MTX OUT
        cdef MTX SELFINV
        OUT = MTX('')
        if n>=0:
            OUT.Data = matpower(self.Data,n)
        else:
            SELFINV = self.__invert__()
            OUT.Data = matpower(SELFINV.Data,-n)
        if OUT.Data:
            OUT._mutable=self._mutable
            return OUT
        else:
            raise RuntimeError, "Something went wrong"

    def __invert__(MTX self):
        """
        M__invert__(): return M^(-1), if defined

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M = MTX(9,[[2,3,4],[5,6,7],[8,9,0]])
            sage: print ~M    # indirect doctest
            [0 0 4]
            [8 3 6]
            [2 2 2]
            
        TESTS::

            sage: M = MTX(9,[[2,3,4],[5,6,7]])
            sage: print M^(-1)    # indirect doctest
            Traceback (most recent call last):
            ...
            ArithmeticError: This MTX matrix is not invertible
            sage: M = MTX(9,[[2,3,4],[2,3,4],[8,9,0]])
            sage: print ~M
            Traceback (most recent call last):
            ...
            ArithmeticError: This MTX matrix is not invertible
        
        """
        if (self.ncols()==0):
            raise NotImplementedError, "The matrix must not be empty"
        cdef MTX OUT
        OUT = MTX('',self._mutable)
        #if self.nullity()==0:
        _sig_on
        OUT.Data = matinv(self.Data)
        _sig_off
        if OUT.Data:
            return OUT
        else:
            raise ArithmeticError, "This MTX matrix is not invertible"

    cpdef MTX transpose(MTX self):
        """
        Return the transposed matrix

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(9,[[2,3,4],[5,6,7],[8,9,0]])
            sage: print M
            [2 3 4]
            [5 6 7]
            [8 0 0]
            sage: print transpose(M)
            [2 5 8]
            [3 6 0]
            [4 7 0]
        """
        if (self.ncols()==0):
            raise NotImplementedError, "The matrix must not be empty"
        cdef MTX OUT
        OUT = MTX('',self._mutable)
        OUT.Data = mattr(self.Data)
        return OUT

    cpdef int order(MTX self):
        """
        Return multiplicative order of self 

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(9,[[2,3,4],[5,6,7],[8,9,0]])
            sage: order(M)
            728
        """
        if (self.ncols()==0):
            raise NotImplementedError, "The matrix must not be empty"
        if (self.Data.nor <> self.Data.noc):
            raise NotImplementedError, "only defined for square matrices"
        cdef int o = matorder(self.Data)
        if o==-1:
            raise ArithmeticError, "order too large"
        else:
            return o

    def stack(self, *L):
        """
        Stack one matrix over one or several others, provided the number of columns coincides

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(5,[[4,3,2,1],[3,2,1,0]])
            sage: N=MTX(5,[[1,1,2,2],[3,3,4,4]])
            sage: print M.stack(N)
            [4 3 2 1]
            [3 2 1 0]
            [1 1 2 2]
            [3 3 4 4]
            sage: print M.stack(N,M)
            [4 3 2 1]
            [3 2 1 0]
            [1 1 2 2]
            [3 3 4 4]
            [4 3 2 1]
            [3 2 1 0]
        """
        cdef MTX N
        cdef MTX OUT
        cdef PTR cur
        cdef int totalrows = self.Data.nor
        for N in L:
            if self.Data.fl!=N.Data.fl:
                raise TypeError, "Matrices must be defined over the same field"
            if self.Data.noc!=N.Data.noc:
                raise RuntimeError, "Matrices must have the same number of columns"
            totalrows += N.Data.nor
        OUT = MTX(self.Data.fl, totalrows, self.Data.noc)
        zsetfield(self.Data.fl)
        zsetlen(self.Data.noc)
        cur = OUT.Data.d
        if not copyrows(cur,self.Data.d,self.Data.nor):
            raise MemoryError, "Error copying first matrix"
        zadvance(&(cur),self.Data.nor)
        for i in range(len(L)):
            N = L[i]
            if not copyrows(cur,N.Data.d,N.Data.nor):
                raise MemoryError, "Error copying additional matrix number %d"%(i)
            zadvance(&(cur),N.Data.nor)
        return OUT


###################
## Gauss algorithm

    cpdef MTX normalized(MTX self):
        """
        Return normalized copy of self,  i.e., all leading coefficients are 1

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(9,[[2,3,4],[5,6,7],[8,9,0]])
            sage: print M.normalized()
            [1 6 8]
            [1 8 2]
            [1 0 0]
        """
        if self.Data.fl==2:
            return self.__copy__()
        cdef MTX OUT = self.__copy__() # MTX(self.Data.fl, self.Data.nor, self.Data.noc)
        cdef int i #,f,j
        cdef PTR p = OUT.Data.d
        cdef piv_t P
        cdef FEL f
        for i from 0<=i<self.Data.nor:
            P = _zfindpiv_(p) # f,j = self[i].lead()
            if P.m:
                if P.m!=1:
                    f = tmultinv[zitof(P.m)]
                    zmulrow(p,f) # OUT[i] = self[i]/P.m
            zsteprow(&(p))
        return OUT

    cpdef MTX echelon(MTX self):
        """
        Return a normalized semi echelon form of self

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(9,[[1,2,3],[4,5,6],[7,8,9]])
            sage: print M.echelon()
            [1 2 3]
            [0 1 5]
        """
        if (self.ncols()==0):
            raise NotImplementedError, "The matrix must not be empty"
        cdef MTX OUT
        # cdef MTX tmp
        cdef int i# ,f,j
        OUT = MTX('',mutable=True)
        _sig_on
        OUT.Data = echelon(self.Data)
        _sig_off
        cdef PTR p = OUT.Data.d
        cdef piv_t P
        cdef FEL f
        if self.Data.fl != 2:
            for i from 0<=i<OUT.Data.nor:
                P = _zfindpiv_(p) # f,j = OUT[i].lead()
                if P.m:
                    if P.m!=1:
                        f = tmultinv[zitof(P.m)]
                        zmulrow(p,f) # OUT[i] = OUT[i]/f
                zsteprow(&(p))
        if not (self._mutable):
            OUT._mutable=False
        return OUT

    cpdef MTX full_echelon(MTX self):
        """
        Return a full normalized echelon form of self

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(9,[[1,2,3],[4,5,6],[7,8,9]])
            sage: print M.full_echelon()
            [0 1 5]
            [1 0 8]
        """
        cdef MTX tmp = self.echelon()
        cdef MTX OUT = MTX(self.Data.fl,tmp.Data.nor,self.Data.noc)
        cdef int i
        #for i from 0<= i < tmp.Data.nor:
        #    OUT[i]=tmp[tmp.Data.nor-i-1]
        cdef PTR src = tmp.Data.d
        cdef PTR tgt = OUT.Data.d
        zadvance(&(tgt),(tmp.Data.nor)-1)
        for i from 0<= i < tmp.Data.nor:
            #zcpyrow(tgt,src,zsize(1))
            zmoverow(tgt,src)
            zsteprow(&(src))
            zbackrow(&(tgt))
        return OUT.echelon()

    cpdef list pivots(MTX self):
        """
        Return the list of pivots of self

        ASSUMPTION:

        The matrix is in echelong form without zero rows

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(5,[[4,4,2,1],[3,3,1,0]])
            sage: print M.echelon()
            [1 1 3 4]
            [0 0 1 4]
            sage: M.echelon().pivots()
            [0, 2]
        """
        cdef list L = []
        cdef int i,j
        if not self.Data:
            return L
        cdef PTR p = self.Data.d
        cdef piv_t P
        for i from 0<=i<self.Data.nor:
            P = _zfindpiv_(p) # f,j = self[i].lead()
            L.append(P.i)
            zsteprow(&(p))
        return L

    cpdef int nullity(MTX self):
        """
        Return the nullity of a matrix

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(5,[[4,3,2,1],[3,2,1,0]])
            sage: M.nullity()
            2
        """
        if (self.ncols()==0):
            raise NotImplementedError, "The matrix must not be empty"
        return nullity(self.Data)

    cpdef MTX nullspace(MTX self):
        """
        Return the null space of a matrix ``M``, i.e., ``M.nullspace()*M`` vanishes

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(5,[[4,3], [3,2], [2,1], [1,0]])
            sage: print M.nullspace()
            [1 3 1 0]
            [0 1 3 1]
            sage: print M.nullspace()*M
            [0 0]
            [0 0]
        """
        if (self.ncols()==0):
            raise NotImplementedError, "The matrix must not be empty"
        cdef MTX OUT
        OUT = MTX('',mutable=self._mutable)
        _sig_on
        OUT.Data = nullspace(self.Data)
        _sig_off
        return OUT

    cpdef tuple lead(self):
        """
        ``(f,i) = M.lead()`` `\\iff` ``f=M[0,i]`` is the first non-zero coefficient in the first row of M.
        
        If the first row of M has no non-zero entry then ``f==0`` and ``i`` is the number of columns

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M = MTX(5,[[0,2,4,1]])
            sage: M.lead()
            (2, 1)
            sage: N = MTX(5, 3,15)
            sage: N.lead()
            (0, 15)
            
        """
        if (self.ncols()==0):
            raise NotImplementedError, "The matrix must not be empty"
        zsetfield(self.Data.fl)
        zsetlen(self.Data.noc)
        cdef piv_t P = _zfindpiv_(self.Data.d)
        return (P.m, P.i)

#######################
## String representations
    def __repr__(self):
        """
        Return a short description of an MTX matrix

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(5,[[4,3,2,1],[3,2,1,0]])
            sage: M   # indirect doctest
            (2 x 4) MTX matrix over GF(5)
        """
        if not self.Data:
            return 'Empty MTX matrix'
        return '(%s x %s) MTX matrix over GF(%s)'%(self.Data.nor, self.Data.noc, self.Data.fl)
    
    def __str__(self):
        """
        Return a string showing the contents of an MTX matrix

        EXAMPLES::

            sage: from pGroupCohomology.mtx import MTX
            sage: M=MTX(125,[[40,113,27,1],[23,112,1,10]])
            sage: print M   # indirect doctest
            [ 40 113  27   1]
            [ 23 112   1  10]
        """
        cdef long i,j
        if not self.Data:
            return '[]'
        nc=self.Data.noc
        nr=self.Data.nor
        zsetfield(self.Data.fl)
        fln=len(str(self.Data.fl))
        cdef list matL = self._matlist_()
        return "\n".join(["["+" ".join([str(el).rjust(fln) for el in matL[i]])+"]" \
                                   for i in range(nr)])
